/*
  PushIt - A simple Sokoban solving game
  Copyright (C) 2006  Tim Teulings

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream>

#include <Lum/Base/String.h>

#include <Lum/Button.h>
#include <Lum/ButtonRow.h>
#include <Lum/Dialog.h>
#include <Lum/Label.h>
#include <Lum/LED.h>
#include <Lum/Menu.h>
#include <Lum/Panel.h>
#include <Lum/Space.h>
#include <Lum/Table.h>
#include <Lum/Text.h>
#include <Lum/TextValue.h>
#include <Lum/WindowGroup.h>

#include <Lum/Dlg/About.h>
#include <Lum/Dlg/Msg.h>

#include <Lum/Model/Action.h>
#include <Lum/Model/Selection.h>
#include <Lum/Model/Table.h>

#include <Lum/OS/Probe.h>

#include "config.h"
#include "Configuration.h"
#include "GameArea.h"

static Lum::AppInfo info;

class SelectCollection : public Lum::Dialog
{
private:
  Lum::Model::ActionRef              okAction;
  Lum::Model::ListTableRef           list;
  Lum::Model::SingleLineSelectionRef selection;
  Collection                         *result;

public:
  SelectCollection()
  : okAction(new Lum::Model::Action),
    list(new Lum::Model::ListTable(2)),
    selection(new Lum::Model::SingleLineSelection),
    result(NULL)
  {
    okAction->Disable();

    AttachModel(okAction);
    AttachModel(selection);

    for (size_t i=0; i<Collection::GetCollectionCount(); ++i) {
      Lum::Model::ListTable::StdEntry *entry;
      Collection                      *collection=Collection::GetCollection(i);

      entry=new Lum::Model::ListTable::StdEntry(list);
      entry->SetString(1,collection->GetTitle());
      if (collection->IsFinished()) {
        entry->SetString(2,L"finished");
      }
      else {
        entry->SetString(2,Lum::Base::NumberToWString(collection->GetCurrent()+1)+
                         L"/"+
                         Lum::Base::NumberToWString(collection->GetLevelCount()));
      }
      list->Append(entry);
    }
  }

  ~SelectCollection()
  {
    UnattachModel(selection);
    UnattachModel(okAction);
  }

  void PreInit()
  {
    Lum::Button           *button;
    Lum::ButtonRow        *bRow;
    Lum::Panel            *panel;
    Lum::Table            *table;
    Lum::WindowGroup      *wGroup;
    Lum::Model::HeaderRef headerModel;

    wGroup=new Lum::WindowGroup();
    wGroup->SetFlex(true,true);
    wGroup->SetWidth(Lum::Base::Size::modePixel,(GetWindow()->GetParent()->GetWidth()*80)/100);
    wGroup->SetHeight(Lum::Base::Size::modePixel,(GetWindow()->GetParent()->GetHeight()*80)/100);

    panel=new Lum::VPanel();
    panel->SetFlex(true,true);

    headerModel=new Lum::Model::HeaderImpl();
    headerModel->AddColumn(L"Collection",Lum::Base::Size::stdCharWidth,25);
    headerModel->AddColumn(L"Current level",Lum::Base::Size::stdCharWidth,13);

    table=new Lum::Table();
    table->SetFlex(true,true);
    table->SetShowHeader(true);
    table->SetModel(list);
    table->SetHeaderModel(headerModel);
    table->SetSelection(selection);
    table->SetDoubleClickAction(okAction);
    panel->Add(table);

    panel->Add(new Lum::VSpace());

    bRow=new Lum::ButtonRow();
    bRow->SetFlex(true,false);
    panel->Add(bRow);

    button=new Lum::Button();
    button->SetFlex(true,true);
    button->SetModel(okAction);
    button->SetType(Lum::Button::typeCommit);
    button->SetText(L"_Ok");
    bRow->Add(button);

    button=new Lum::Button();
    button->SetFlex(true,true);
    button->SetModel(GetClosedAction());
    button->SetType(Lum::Button::typeCancel);
    button->SetText(L"_Cancel");
    bRow->Add(button);

    wGroup->SetMain(panel);

    SetTop(wGroup);

    Dialog::PreInit();
  }

  void Resync(Lum::Base::Model *model, const Lum::Base::ResyncMsg& msg)
  {
    if (model==okAction && okAction->IsFinished()) {
      if (selection->HasSelection()) {
        result=Collection::GetCollection(selection->GetLine()-1);
      }

      Exit();
    }
    else if (model==selection) {
      if (selection->HasSelection()) {
        okAction->Enable();
      }
      else {
        okAction->Disable();
      }
    }

    Dialog::Resync(model,msg);
  }

  Collection* GetResult() const
  {
    return result;
  }
};

class SelectLevel : public Lum::Dialog
{
private:
  Collection                         *collection;
  Lum::Model::ActionRef              okAction;
  Lum::Model::ListTableRef           list;
  Lum::Model::SingleLineSelectionRef selection;
  bool                               success;
  size_t                             result;

public:
  SelectLevel(Collection *collection)
  : collection(collection),
    okAction(new Lum::Model::Action),
    list(new Lum::Model::ListTable(1)),
    selection(new Lum::Model::SingleLineSelection),
    success(false),
    result(0)
  {
    okAction->Disable();

    AttachModel(okAction);
    AttachModel(selection);

    for (size_t i=0; i<collection->GetLevelCount() && i<=collection->GetCurrent(); ++i) {
      list->AppendString(collection->GetLevel(i).GetTitle());
    }
  }

  ~SelectLevel()
  {
    UnattachModel(selection);
    UnattachModel(okAction);
  }

  void PreInit()
  {
    Lum::Button      *button;
    Lum::ButtonRow   *bRow;
    Lum::Panel       *panel;
    Lum::Table       *table;
    Lum::WindowGroup *wGroup;

    wGroup=new Lum::WindowGroup();
    wGroup->SetFlex(true,true);
    wGroup->SetWidth(Lum::Base::Size::modePixel,(GetWindow()->GetParent()->GetWidth()*80)/100);
    wGroup->SetHeight(Lum::Base::Size::modePixel,(GetWindow()->GetParent()->GetHeight()*80)/100);

    panel=new Lum::VPanel();
    panel->SetFlex(true,true);

    table=new Lum::Table();
    table->SetFlex(true,true);
    table->SetModel(list);
    table->SetSelection(selection);
    table->SetDoubleClickAction(okAction);
    panel->Add(table);

    panel->Add(new Lum::VSpace());

    bRow=new Lum::ButtonRow();
    bRow->SetFlex(true,false);
    panel->Add(bRow);

    button=new Lum::Button();
    button->SetFlex(true,true);
    button->SetModel(okAction);
    button->SetType(Lum::Button::typeCommit);
    button->SetText(L"_Ok");
    bRow->Add(button);

    button=new Lum::Button();
    button->SetFlex(true,true);
    button->SetModel(GetClosedAction());
    button->SetType(Lum::Button::typeCancel);
    button->SetText(L"_Cancel");
    bRow->Add(button);

    wGroup->SetMain(panel);

    SetTop(wGroup);

    Dialog::PreInit();
  }

  void Resync(Lum::Base::Model *model, const Lum::Base::ResyncMsg& msg)
  {
    if (model==okAction && okAction->IsFinished()) {
      if (selection->HasSelection()) {
        result=selection->GetLine()-1;
        success=true;
      }

      Exit();
    }
    else if (model==selection) {
      if (selection->HasSelection()) {
        okAction->Enable();
      }
      else {
        okAction->Disable();
      }
    }

    Dialog::Resync(model,msg);
  }

  bool Success() const
  {
    return success;
  }

  size_t GetResult() const
  {
    return result;
  }
};

class SelectTheme : public Lum::Dialog
{
private:
  Lum::Model::ActionRef              okAction;
  Lum::Model::ListTableRef           list;
  Lum::Model::SingleLineSelectionRef selection;
  std::wstring                       result;

public:
  SelectTheme()
  : okAction(new Lum::Model::Action),
    list(new Lum::Model::ListTable(2)),
    selection(new Lum::Model::SingleLineSelection),
    result(L"")
  {
    okAction->Disable();

    AttachModel(okAction);
    AttachModel(selection);


    for (size_t i=0; i<themes.size(); ++i) {
      Lum::Model::ListTable::StdEntry *entry;

      entry=new Lum::Model::ListTable::StdEntry(list);
      entry->SetString(1,themes[i].name);
      entry->SetString(2,Lum::Base::NumberToWString(themes[i].boxSize));

      list->Append(entry);
    }
  }

  ~SelectTheme()
  {
    UnattachModel(selection);
    UnattachModel(okAction);
  }

  void PreInit()
  {
    Lum::Button           *button;
    Lum::ButtonRow        *bRow;
    Lum::Panel            *panel;
    Lum::Table            *table;
    Lum::WindowGroup      *wGroup;
    Lum::Model::HeaderRef headerModel;

    wGroup=new Lum::WindowGroup();
    wGroup->SetFlex(true,true);
    wGroup->SetWidth(Lum::Base::Size::modePixel,(GetWindow()->GetParent()->GetWidth()*80)/100);
    wGroup->SetHeight(Lum::Base::Size::modePixel,(GetWindow()->GetParent()->GetHeight()*80)/100);

    panel=new Lum::VPanel();
    panel->SetFlex(true,true);

    headerModel=new Lum::Model::HeaderImpl();
    headerModel->AddColumn(L"Theme name",Lum::Base::Size::stdCharWidth,25);
    headerModel->AddColumn(L"box size",Lum::Base::Size::stdCharWidth,13);

    table=new Lum::Table();
    table->SetFlex(true,true);
    table->SetShowHeader(true);
    table->SetModel(list);
    table->SetHeaderModel(headerModel);
    table->SetSelection(selection);
    table->SetDoubleClickAction(okAction);
    panel->Add(table);

    panel->Add(new Lum::VSpace());

    bRow=new Lum::ButtonRow();
    bRow->SetFlex(true,false);
    panel->Add(bRow);

    button=new Lum::Button();
    button->SetFlex(true,true);
    button->SetModel(okAction);
    button->SetType(Lum::Button::typeCommit);
    button->SetText(L"_Ok");
    bRow->Add(button);

    button=new Lum::Button();
    button->SetFlex(true,true);
    button->SetModel(GetClosedAction());
    button->SetType(Lum::Button::typeCancel);
    button->SetText(L"_Cancel");
    bRow->Add(button);

    wGroup->SetMain(panel);

    SetTop(wGroup);

    Dialog::PreInit();
  }

  void Resync(Lum::Base::Model *model, const Lum::Base::ResyncMsg& msg)
  {
    if (model==okAction && okAction->IsFinished()) {
      if (selection->HasSelection()) {
        result=themes[selection->GetLine()-1].name;
      }

      Exit();
    }
    else if (model==selection) {
      if (selection->HasSelection()) {
        okAction->Enable();
      }
      else {
        okAction->Disable();
      }
    }

    Dialog::Resync(model,msg);
  }

  std::wstring GetResult() const
  {
    return result;
  }
};

class MainDialog : public Lum::Dialog
{
private:
  Lum::Model::ActionRef categoryAction;
  Lum::Model::ActionRef levelAction;
  Lum::Model::ActionRef themeAction;
  Lum::Model::ActionRef restartAction;
  Lum::Model::ActionRef undoAction;
  Lum::Model::ActionRef aboutAction;
  Lum::Model::StringRef titleName;
  Lum::Model::StringRef levelName;
  Lum::Model::StringRef themeName;
  Lum::Model::StringRef moves;
  Lum::Model::StringRef pushes;
  GameArea              *gameArea;
  size_t                level;

public:
  MainDialog()
  : categoryAction(new Lum::Model::Action),
    levelAction(new Lum::Model::Action),
    themeAction(new Lum::Model::Action),
    restartAction(new Lum::Model::Action),
    undoAction(new Lum::Model::Action),
    aboutAction(new Lum::Model::Action),
    titleName(new Lum::Model::String()),
    levelName(new Lum::Model::String()),
    themeName(new Lum::Model::String()),
    moves(new Lum::Model::String()),
    pushes(new Lum::Model::String()),
    gameArea(new GameArea)
  {
    AttachModel(categoryAction);
    AttachModel(levelAction);
    AttachModel(themeAction);
    AttachModel(restartAction);
    AttachModel(undoAction);
    AttachModel(aboutAction);

    AttachModel(gameArea->GetFinishedModel());

    moves->Set(L"0");
    pushes->Set(L"0");
  }

  ~MainDialog()
  {
    UnattachModel(gameArea->GetFinishedModel());

    UnattachModel(aboutAction);
    UnattachModel(undoAction);
    UnattachModel(restartAction);
    UnattachModel(themeAction);
    UnattachModel(levelAction);
    UnattachModel(categoryAction);
  }

  void PreInit()
  {
    Lum::Button      *button;
    Lum::Label       *label;
    Lum::LED         *led;
    Lum::Menu        *menu;
    Lum::MenuStrip   *menuStrip;
    Lum::Panel       *horiz;
    Lum::Panel       *horiz2;
    Lum::Panel       *vert;
    Lum::TextValue   *textValue;
    Lum::WindowGroup *wGroup;

    wGroup=new Lum::WindowGroup();
    wGroup->SetFlex(true,true);

    horiz=new Lum::HPanel();
    horiz->SetFlex(true,true);

    vert=new Lum::VPanel();
    vert->SetFlex(true,true);

    gameArea->SetFlex(true,true);
    vert->Add(gameArea);

    horiz->Add(vert);

    horiz->Add(new Lum::HSpace());

    vert=new Lum::VPanel();
    vert->SetFlex(false,true);

    label=new Lum::Label;

    horiz2=new Lum::HPanel();
    horiz2->SetFlex(true,false);

    textValue=new Lum::TextValue();
    textValue->SetFlex(true,false);
    textValue->SetMinWidth(Lum::Base::Size::stdCharWidth,10);
    textValue->SetModel(titleName);
    horiz2->Add(textValue);

    button=new Lum::Button();
    button->SetModel(categoryAction);
    button->SetText(L"...");
    horiz2->Add(button);

    label->AddLabel(L"Collection:",horiz2);

    horiz2=new Lum::HPanel();
    horiz2->SetFlex(true,false);

    textValue=new Lum::TextValue();
    textValue->SetFlex(true,false);
    textValue->SetMinWidth(Lum::Base::Size::stdCharWidth,10);
    textValue->SetModel(levelName);
    horiz2->Add(textValue);

    button=new Lum::Button();
    button->SetModel(levelAction);
    button->SetText(L"...");
    horiz2->Add(button);

    label->AddLabel(L"Level:",horiz2);

    horiz2=new Lum::HPanel();
    horiz2->SetFlex(true,false);

    textValue=new Lum::TextValue();
    textValue->SetFlex(true,false);
    textValue->SetMinWidth(Lum::Base::Size::stdCharWidth,10);
    textValue->SetModel(themeName);
    horiz2->Add(textValue);

    button=new Lum::Button();
    button->SetModel(themeAction);
    button->SetText(L"...");
    horiz2->Add(button);

    label->AddLabel(L"Theme:",horiz2);

    textValue=new Lum::TextValue();
    textValue->SetFlex(true,false);
    textValue->SetMinWidth(Lum::Base::Size::stdCharWidth,5);
    textValue->SetModel(moves);
    label->AddLabel(L"Moves:",textValue);

    textValue=new Lum::TextValue();
    textValue->SetFlex(true,false);
    textValue->SetMinWidth(Lum::Base::Size::stdCharWidth,5);
    textValue->SetModel(pushes);
    label->AddLabel(L"Pushes:",textValue);

    led=new Lum::LED;
    led->SetModel(gameArea->GetFinishedModel());
    label->AddLabel(L"Finished:",led);

    vert->Add(label);

    vert->Add(new Lum::VSpace());

    button=new Lum::Button();
    button->SetFlex(true,false);
    button->SetModel(restartAction);
    button->SetText(L"_Restart level");
    vert->Add(button);

    vert->Add(new Lum::VSpace());

    button=new Lum::Button();
    button->SetFlex(true,false);
    button->SetModel(undoAction);
    button->SetText(L"_Undo step");
    vert->Add(button);

    horiz->Add(vert);

    wGroup->SetMain(horiz);

    menuStrip=new Lum::MenuStrip();

    menu=new Lum::Menu();
    menu->AddActionItem(L"_Quit",GetClosedAction(),Lum::OS::qualifierControl,L"q");
    menuStrip->AddPullDownMenu(L"_Project",menu);

    menu=new Lum::Menu();
    menu->AddActionItem(L"_Choose _category",categoryAction,Lum::OS::qualifierControl,L"c");
    menu->AddActionItem(L"Choose _level",levelAction,Lum::OS::qualifierControl,L"l");
    menu->AddActionItem(L"Choose _theme",themeAction,Lum::OS::qualifierControl,L"t");
    menu->AddSeparator();
    menu->AddActionItem(L"_Restart level",restartAction,Lum::OS::qualifierControl,L"r");
    menu->AddActionItem(L"_Undo step",restartAction,Lum::OS::qualifierControl,L"u");
    menuStrip->AddPullDownMenu(L"_Game",menu);

    menu=new Lum::Menu();
    menu->AddActionItem(L"_About",aboutAction);
    menuStrip->AddPullDownMenu(L"_Help",menu);

    wGroup->SetMenuStrip(menuStrip);

    SetTop(wGroup);

    Dialog::PreInit();
  }

  void SetLevel(size_t level)
  {
    assert(level<collection->GetLevelCount());
    this->level=level;

    gameArea->SetSokoban(collection->GetLevel(level).sokoban);
    levelName->Set(collection->GetLevel(level).GetTitle());
  }

  void SetCollection(Collection *collection)
  {
    assert(collection!=NULL);

    ::collection=collection;
    titleName->Set(collection->GetTitle());

    if (collection->IsFinished()) {
      SetLevel(0);
    }
    else {
      SetLevel(collection->GetCurrent());
    }
  }

  void Resync(Lum::Base::Model *model, const Lum::Base::ResyncMsg& msg)
  {
    if (model==GetOpenedAction() && GetOpenedAction()->IsFinished()) {
      if (!LoadLevels()) {
        Lum::Dlg::Msg::ShowOk(this,
                              L"Cannot find any levels!",
                              L"The program was not able to find any sokoban levels!\n"
                              L"\n"
                              L"This is normally the sign of an uncomplete or corrupt\n"
                              L"installation. Please check your installation and reinstall\n"
                              L"the program, if in doubt.\n"
                              L"\n"
                              L"You cannot use the program and thus it will now exit!");

        Exit();
        return;
      }

      LoadThemes();

      if (!LoadConfig()) {
        Lum::Dlg::Msg::ShowOk(this,
                              L"No configuration file found!",
                              L"The program was not able to find its configuration file.\n"
                              L"\n"
                              L"If this is the first start of your program, this is\n"
                              L"OK. The program will generate a configuration file if\n"
                              L"it quits.\n"
                              L"If this is not your first use of this program, you may not\n"
                              L"have solved any levels up to now, have not correctly quit\n"
                              L"the program, or have deleted the history of solved levels!");
      }

      gameArea->EvaluateThemeName();
      themeName->Set(::themeName);
      if (collection!=NULL) {
        SetCollection(collection);
      }
      else {
        SetCollection(Collection::GetCollection(0));
      }
    }
    else if (model==categoryAction && categoryAction->IsFinished()) {
      SelectCollection *dialog=new SelectCollection();

      dialog->SetParent(this);
      if (dialog->Open()) {
        dialog->SetExitAction(dialog->GetClosedAction());
        dialog->EventLoop();
        dialog->Close();
      }

      if (dialog->GetResult()!=NULL) {
        SetCollection(dialog->GetResult());
      }

      delete dialog;
    }
    else if (model==levelAction && levelAction->IsFinished()) {
      SelectLevel *dialog=new SelectLevel(collection);

      dialog->SetParent(this);
      if (dialog->Open()) {
        dialog->SetExitAction(dialog->GetClosedAction());
        dialog->EventLoop();
        dialog->Close();
      }

      if (dialog->Success()) {
        SetLevel(dialog->GetResult());
      }

      delete dialog;
    }
    else if (model==themeAction && themeAction->IsFinished()) {
      SelectTheme *dialog=new SelectTheme();

      dialog->SetParent(this);
      if (dialog->Open()) {
        dialog->SetExitAction(dialog->GetClosedAction());
        dialog->EventLoop();
        dialog->Close();
      }

      if (!dialog->GetResult().empty()) {
        ::themeName=dialog->GetResult();
        themeName->Set(dialog->GetResult());
        gameArea->EvaluateThemeName();
      }

      delete dialog;
    }
    else if (model==restartAction && restartAction->IsFinished()) {
      SetLevel(level);
    }
    else if (model==undoAction && undoAction->IsFinished()) {
      gameArea->Undo();
    }
    else if (model==aboutAction && aboutAction->IsFinished()) {
      Lum::Dlg::About::Show(this,info);
    }
    else if (model==gameArea->GetFinishedModel() && gameArea->GetFinishedModel()->Get()) {
      if (collection->GetCurrent()<level+1) {
        collection->SetCurrent(level+1);
      }
      if (!collection->IsFinished()) {
        SetLevel(level+1);
      }
    }

    Dialog::Resync(model,msg);
  }
};

int main(int argc, char* argv[])
{
  info.SetProgram(Lum::Base::StringToWString(PACKAGE_NAME));
  info.SetVersion(Lum::Base::StringToWString(PACKAGE_VERSION));
  info.SetDescription(L"Move everything to its place...");
  info.SetAuthor(L"Tim Teulings");
  info.SetContact(L"Tim Teulings <tim@teulings.org>");
  info.SetCopyright(L"(c) 2006, Tim Teulings");
  info.SetLicense(L"GNU Public License");

  if (Lum::OS::prober->Open(L"PushIt",argc,argv)) {
    Lum::Dialog *window;

    window=new MainDialog();

    if (window->Open()) {
      window->SetExitAction(window->GetClosedAction());
      window->EventLoop();
      window->Close();
    }

    delete window;

    SaveConfig();
    FreeConfig();

    Lum::OS::display->Close();
  }
}
